from collections.abc import Iterable
from typing import Any

from motor.motor_asyncio import AsyncIOMotorClient
from pymongo.errors import PyMongoError

from pydongo.drivers.base import AbstractAsyncMongoDBDriver
from pydongo.expressions.index import IndexExpression


class MotorMongoDBDriver(AbstractAsyncMongoDBDriver):
    """Default asynchronous MongoDB driver using Motor (async wrapper for PyMongo).

    This driver connects to a real MongoDB instance and provides async methods
    for insert, update, query, and delete operations.
    """

    def __init__(self, connection_string: str, database_name: str):
        """Initialize the driver with a MongoDB URI and target database name.

        Args:
            connection_string (str): MongoDB connection URI.
            database_name (str): Name of the database to use.
        """
        self.connection_string = connection_string
        self.database_name = database_name
        self.client: AsyncIOMotorClient | None = None  # type: ignore  # noqa: PGH003

    async def connect(self) -> bool:
        """Asynchronously establish a connection to the MongoDB server.

        Returns:
            bool: True if the connection is successful.

        Raises:
            RuntimeError: If unable to connect to MongoDB.
        """
        try:
            self.client = AsyncIOMotorClient(self.connection_string)
            self.db = self.client[self.database_name]
            return True
        except PyMongoError as e:
            raise RuntimeError(f"Failed to connect to MongoDB: {e}") from e

    async def close(self) -> None:
        """Asynchronously close the MongoDB connection."""
        if self.client:
            self.client.close()

    async def insert_one(self, collection: str, document: dict[str, Any]) -> dict[str, Any]:
        """Insert a single document into the specified collection.

        Args:
            collection (str): Name of the collection.
            document (dict): Document to insert.

        Returns:
            dict: Result including the inserted document ID.
        """
        result = await self.db[collection].insert_one(document)
        return {"inserted_id": str(result.inserted_id)}

    async def insert_many(self, collection: str, documents: list[dict[str, Any]]) -> dict[str, Any]:
        """Insert multiple documents into the specified collection.

        Args:
            collection (str): Name of the collection.
            documents (list): Documents to insert.

        Returns:
            dict: Result including inserted document IDs.
        """
        result = await self.db[collection].insert_many(documents)
        return {"inserted_ids": [str(_id) for _id in result.inserted_ids]}

    async def find_one(self, collection: str, query: dict[str, Any]) -> dict[str, Any] | None:
        """Find a single document that matches the query.

        Args:
            collection (str): Name of the collection.
            query (dict): MongoDB query object.

        Returns:
            dict or None: The matching document or None.
        """
        return await self.db[collection].find_one(query)

    async def find_many(
        self,
        collection: str,
        query: dict[str, Any],
        sort_criteria: dict[str, int],
        offset: int | None = None,
        limit: int | None = None,
    ) -> Iterable[dict[str, Any]]:
        """Find multiple documents matching the query.

        Args:
            collection (str): Name of the collection.
            query (dict): MongoDB filter query.
            sort_criteria (dict): Sort order (e.g., {"created_at": -1}).
            offset (Union[int, None]): Number of documents to skip.
            limit (Union[int, None]): Max number of documents to return.

        Returns:
            Iterable[dict]: A cursor of matching documents.
        """
        cursor = self.db[collection].find(query)

        if sort_criteria:
            cursor = cursor.sort(sort_criteria)

        if offset is not None:
            cursor = cursor.skip(offset)

        if limit is not None:
            cursor = cursor.limit(limit)

        return cursor  # type: ignore[return-value]

    async def update_one(
        self,
        collection: str,
        query: dict[str, Any],
        update: dict[str, Any],
        upsert: bool = False,
    ) -> dict[str, Any]:
        """Update a single document that matches the filter.

        Args:
            collection (str): Name of the collection.
            query (dict): MongoDB filter.
            update (dict): MongoDB update expression.
            upsert (bool): If True, create the document if it does not exist.

        Returns:
            dict: Result of the update operation.
        """
        result = await self.db[collection].update_one(query, update, upsert=upsert)
        return {
            "matched_count": result.matched_count,
            "modified_count": result.modified_count,
            "upserted_id": str(result.upserted_id) if result.upserted_id else None,
        }

    async def delete_one(self, collection: str, query: dict[str, Any]) -> dict[str, Any]:
        """Delete a single document matching the query.

        Args:
            collection (str): Name of the collection.
            query (dict): MongoDB filter.

        Returns:
            dict: Result including count of deleted documents.
        """
        result = await self.db[collection].delete_one(query)
        return {"deleted_count": result.deleted_count}

    async def count(self, collection: str, query: dict[str, Any]) -> int:
        """Count documents matching the given filter.

        Args:
            collection (str): Name of the collection.
            query (dict): MongoDB filter.

        Returns:
            int: Number of matching documents.
        """
        return await self.db[collection].count_documents(query)

    async def exists(self, collection: str, query: dict[str, Any]) -> bool:
        """Check if at least one document exists that matches the filter.

        Args:
            collection (str): Name of the collection.
            query (dict): MongoDB filter.

        Returns:
            bool: True if a matching document exists.
        """
        return await self.db[collection].count_documents(query, limit=1) > 0

    async def create_index(self, collection: str, index: tuple[IndexExpression]) -> None:
        """Create an index on a collection in the MongoDB Database (Async version using Motor).

        Args:
            collection (str): The collection name.
            index (tuple[IndexExpression]):
                A tuple of IndexExpression objects representing the index to create.
                NOTE: Multiple elements in the tuple indicate a single compound index,
                not multiple separate indexes.
        """
        index_key: list[tuple[str, Any]] = []
        final_kwargs = {}

        for expr in index:
            index_key.extend(expr.serialize().items())
            final_kwargs.update(expr.build_kwargs())

        await self.db[collection].create_index(index_key, **final_kwargs)

    async def update_many(self, collection: str, query: dict[str, Any], update: dict[str, Any]) -> dict[str, Any]:
        """Update multiple documents matching the filter.

        Args:
            collection (str): Name of the collection.
            query (dict): MongoDB filter.
            update (dict): MongoDB update expression.

        Returns:
            dict: Result of the update operation.
        """
        result = await self.db[collection].update_many(query, update)
        return {
            "matched_count": result.matched_count,
            "modified_count": result.modified_count,
            "upserted_id": str(result.upserted_id) if result.upserted_id else None,
        }



import uuid
from collections import defaultdict
from copy import deepcopy
from typing import Any

from pydongo.drivers.base import AbstractAsyncMongoDBDriver, AbstractSyncMongoDBDriver
from pydongo.expressions.index import IndexExpression


class MockMongoDBDriver(AbstractSyncMongoDBDriver):
    """In-memory mock implementation of the AbstractSyncMongoDBDriver.

    This mock driver mimics MongoDB behavior for unit testing without requiring
    a real database connection. Data is stored in-memory in Python dictionaries.
    """

    def __init__(self, connection_string: str = "", database_name: str = "mockdb"):
        """Initialize the mock driver with an empty in-memory store.

        Args:
            connection_string (str): Ignored.
            database_name (str): Name of the fake database.
        """
        self.connection_string = connection_string
        self.database_name = database_name
        self._collections: dict[str, list[dict[str, Any]]] = {}
        self.indexes: defaultdict[str, Any] = defaultdict(list)

    def connect(self) -> bool:
        """Simulate a successful connection.

        Returns:
            bool: Always True.
        """
        return True

    def close(self) -> None:
        """Close the connection (noop for the mock driver)."""

    def insert_one(self, collection: str, document: dict[str, Any]) -> dict[str, Any]:
        """Insert a document into the in-memory collection.

        Args:
            collection (str): Name of the collection.
            document (dict): Document to insert.

        Returns:
            dict: Insert result including generated `_id`.
        """
        document = deepcopy(document)
        document["_id"] = str(uuid.uuid4())
        self._collections.setdefault(collection, []).append(document)
        return {"inserted_id": document["_id"]}

    def insert_many(self, collection: str, documents: list[dict[str, Any]]) -> dict[str, Any]:
        """Insert multiple documents into the in-memory collection.

        Args:
            collection (str): Name of the collection.
            documents (list): List of documents to insert.

        Returns:
            dict: Insert result with list of `_id`s.
        """
        inserted_ids = []
        for doc in documents:
            inserted = self.insert_one(collection, doc)
            inserted_ids.append(inserted["inserted_id"])
        return {"inserted_ids": inserted_ids}

    def find_one(self, collection: str, query: dict[str, Any]) -> dict[str, Any] | None:
        """Find a single document matching the query.

        Args:
            collection (str): Name of the collection.
            query (dict): MongoDB-style filter.

        Returns:
            dict or None: Matching document or None if not found.
        """
        for doc in self._collections.get(collection, []):
            if all(doc.get(k) == v for k, v in query.items()):
                return deepcopy(doc)
        return None

    def find_many(
        self,
        collection: str,
        query: dict[str, Any],
        sort_criteria: dict[str, Any],  # noqa: ARG002
        offset: int | None = None,
        limit: int | None = None,
    ) -> list[dict[str, Any]]:
        """Find multiple documents matching the query.

        Args:
            collection (str): Name of the collection.
            query (dict): MongoDB-style filter.
            sort_criteria (dict): Ignored in the mock.
            offset (Union[int, None]): Skip the first N results.
            limit (Union[int, None]): Limit the number of results returned.

        Returns:
            list: List of matching documents.
        """
        result = [
            deepcopy(doc)
            for doc in self._collections.get(collection, [])
            if all(doc.get(k) == v for k, v in query.items())
        ]
        return result[offset:limit] if limit else result

    def update_one(
        self,
        collection: str,
        query: dict[str, Any],
        update: dict[str, Any],
        upsert: bool = False,  # noqa: ARG002
    ) -> dict[str, Any]:
        """Update a single document matching the query.

        Args:
            collection (str): Name of the collection.
            query (dict): Filter query.
            update (dict): Update operations (only supports "$set").
            upsert (bool): Whether to insert a new document if no match is found.

        Returns:
            dict: Update result with matched and modified count.
        """
        for doc in self._collections.get(collection, []):
            if all(doc.get(k) == v for k, v in query.items()):
                doc.update(update.get("$set", {}))
                return {"matched_count": 1, "modified_count": 1, "upserted_id": None}
        return {"matched_count": 0, "modified_count": 0, "upserted_id": None}

    def delete_one(self, collection: str, query: dict[str, Any]) -> dict[str, Any]:
        """Delete the first document matching the query.

        Args:
            collection (str): Name of the collection.
            query (dict): Filter query.

        Returns:
            dict: Delete result.
        """
        docs = self._collections.get(collection, [])
        for i, doc in enumerate(docs):
            if all(doc.get(k) == v for k, v in query.items()):
                del docs[i]
                return {"deleted_count": 1}
        return {"deleted_count": 0}

    def count(self, collection: str, query: dict[str, Any]) -> int:
        """Count documents that match the query.

        Args:
            collection (str): Name of the collection.
            query (dict): Filter query.

        Returns:
            int: Number of matching documents.
        """
        return sum(1 for doc in self._collections.get(collection, []) if all(doc.get(k) == v for k, v in query.items()))

    def exists(self, collection: str, query: dict[str, Any]) -> bool:
        """Check if at least one document matches the query.

        Args:
            collection (str): Name of the collection.
            query (dict): Filter query.

        Returns:
            bool: True if at least one document matches.
        """
        return any(all(doc.get(k) == v for k, v in query.items()) for doc in self._collections.get(collection, []))

    def create_index(self, collection: str, index: tuple[IndexExpression]) -> None:
        """Create an index on a collection in the MongoDB Database.

        Args:
            collection (str): The collection name.
            index (tuple[IndexExpression]):
                A tuple of IndexExpression objects representing the index to create
                NOTE: Multiple elements in tuple indicate a single compound index
                not multiple indexes

        """
        self.indexes[collection].append(index)

    def update_many(self, collection: str, query: dict[str, Any], update: dict[str, Any]) -> dict[str, Any]:
        """Update multiple documents matching the query."""
        return {
            "collection": collection,
            "query": query,
            "update": update,
        }


class MockAsyncMongoDBDriver(AbstractAsyncMongoDBDriver):
    """In-memory mock implementation of the AbstractAsyncMongoDBDriver.

    This async mock mirrors the behavior of MongoDB using native async/await
    and an in-memory store. Ideal for use in async unit tests.
    """

    def __init__(self, connection_string: str = "", database_name: str = "mockdb"):
        """Initialize the async mock with an in-memory store.

        Args:
            connection_string (str): Ignored.
            database_name (str): Mock database name.
        """
        self.connection_string = connection_string
        self.database_name = database_name
        self._collections: dict[str, list[dict[str, Any]]] = {}
        self.indexes: defaultdict[str, Any] = defaultdict(list)

    async def connect(self) -> bool:
        """Simulate async connection success.

        Returns:
            bool: Always True.
        """
        return True

    async def close(self) -> None:
        """Simulate async close (noop)."""

    async def insert_one(self, collection: str, document: dict[str, Any]) -> dict[str, Any]:
        """Asynchronously insert a document into the mock store.

        Args:
            collection (str): Collection name.
            document (dict): Document to insert.

        Returns:
            dict: Insert result with `_id`.
        """
        document = deepcopy(document)
        document["_id"] = str(uuid.uuid4())
        self._collections.setdefault(collection, []).append(document)
        return {"inserted_id": document["_id"]}

    async def insert_many(self, collection: str, documents: list[dict[str, Any]]) -> dict[str, Any]:
        """Asynchronously insert multiple documents.

        Args:
            collection (str): Collection name.
            documents (list): List of documents.

        Returns:
            dict: Insert result with `_id`s.
        """
        inserted_ids = []
        for doc in documents:
            inserted = await self.insert_one(collection, doc)
            inserted_ids.append(inserted["inserted_id"])
        return {"inserted_ids": inserted_ids}

    async def find_one(self, collection: str, query: dict[str, Any]) -> dict[str, Any] | None:
        """Asynchronously find a single matching document.

        Args:
            collection (str): Collection name.
            query (dict): Filter.

        Returns:
            dict or None: Matching document or None.
        """
        for doc in self._collections.get(collection, []):
            if all(doc.get(k) == v for k, v in query.items()):
                return deepcopy(doc)
        return None

    async def find_many(
        self,
        collection: str,
        query: dict[str, Any],
        sort_criteria: dict[str, Any],  # noqa: ARG002
        offset: int | None = None,
        limit: int | None = None,
    ) -> list[dict[str, Any]]:
        """Asynchronously return a list of matching documents.

        Args:
            collection (str): Collection name.
            query (dict): Filter.
            sort_criteria (dict): Ignored.
            offset (Union[int, None]): Skip N docs.
            limit (Union[int, None]): Limit docs returned.

        Returns:
            list: Matching documents.
        """
        result = [
            deepcopy(doc)
            for doc in self._collections.get(collection, [])
            if all(doc.get(k) == v for k, v in query.items())
        ]
        return result[offset:limit] if limit else result

    async def update_one(
        self,
        collection: str,
        query: dict[str, Any],
        update: dict[str, Any],
        upsert: bool = False,  # noqa: ARG002
    ) -> dict[str, Any]:
        """Asynchronously update a single matching document.

        Args:
            collection (str): Collection name.
            query (dict): Filter.
            update (dict): Update expression.
            upsert (bool): Whether to insert if no match found.

        Returns:
            dict: Update result.
        """
        for doc in self._collections.get(collection, []):
            if all(doc.get(k) == v for k, v in query.items()):
                doc.update(update.get("$set", {}))
                return {"matched_count": 1, "modified_count": 1, "upserted_id": None}
        return {"matched_count": 0, "modified_count": 0, "upserted_id": None}

    async def delete_one(self, collection: str, query: dict[str, Any]) -> dict[str, Any]:
        """Asynchronously delete a single matching document.

        Args:
            collection (str): Collection name.
            query (dict): Filter.

        Returns:
            dict: Delete result.
        """
        docs = self._collections.get(collection, [])
        for i, doc in enumerate(docs):
            if all(doc.get(k) == v for k, v in query.items()):
                del docs[i]
                return {"deleted_count": 1}
        return {"deleted_count": 0}

    async def count(self, collection: str, query: dict[str, Any]) -> int:
        """Asynchronously count matching documents.

        Args:
            collection (str): Collection name.
            query (dict): Filter.

        Returns:
            int: Number of matches.
        """
        return sum(1 for doc in self._collections.get(collection, []) if all(doc.get(k) == v for k, v in query.items()))

    async def exists(self, collection: str, query: dict[str, Any]) -> bool:
        """Asynchronously check if any document matches the filter.

        Args:
            collection (str): Collection name.
            query (dict): Filter.

        Returns:
            bool: True if at least one match exists.
        """
        return any(all(doc.get(k) == v for k, v in query.items()) for doc in self._collections.get(collection, []))

    async def create_index(self, collection: str, index: tuple[IndexExpression]) -> None:
        """Create an index on a collection in the MongoDB Database.

        Args:
            collection (str): The collection name.
            index (tuple[IndexExpression]):
                A tuple of IndexExpression objects representing the index to create
                NOTE: Multiple elements in tuple indicate a single compound index
                not multiple indexes

        """
        self.indexes[collection].append(index)

    async def update_many(self, collection: str, query: dict[str, Any], update: dict[str, Any]) -> dict[str, Any]:
        """Mock async update multiple documents matching the query."""
        return {
            "collection": collection,
            "query": query,
            "update": update,
        }



from enum import Enum, IntEnum
from typing import Any

from pydongo.expressions.base import BaseExpression
from pydongo.expressions.filter import CollectionFilterExpression


class IndexSortOrder(IntEnum):
    """Enum representing index sort order in MongoDB."""

    ASCENDING = 1
    DESCENDING = -1


class IndexType(str, Enum):
    """Enum representing different types of MongoDB indexes."""

    TEXT = "text"
    HASHED = "hashed"
    TWO_DIMENSIONAL = "2d"
    TWO_DIMENSIONAL_SPHERE = "2dsphere"


class CollationStrength(IntEnum):
    """Enum representing the strength levels for string comparison in MongoDB collation."""

    PRIMARY = 1  # Only base characters (e.g., "a" == "A", "é" == "e")
    SECONDARY = 2  # Case-insensitive, accent-sensitive (e.g., "a" == "A", "é" != "e")
    TERTIARY = 3  # Case- and accent-sensitive (e.g., "a" != "A", "é" != "e")
    QUATERNARY = 4  # Includes case, accent, punctuation (e.g., "$a" != "a")
    IDENTICAL = 5  # Everything is compared (e.g., code point level)

    def description(self) -> str:
        """Provides a human-readable description of the collation strength level.

        Returns:
            str: A description of the collation strength level.
        """
        return {
            CollationStrength.PRIMARY: "Base characters only (case-insensitive, accent-insensitive)",
            CollationStrength.SECONDARY: "Accent-sensitive, case-insensitive",
            CollationStrength.TERTIARY: "Case- and accent-sensitive (default)",
            CollationStrength.QUATERNARY: "Includes case, accent, and punctuation",
            CollationStrength.IDENTICAL: "Full Unicode comparison, includes code points",
        }[self]


class IndexExpression(BaseExpression):
    """Represents a MongoDB index definition."""

    def __init__(  # noqa: PLR0913
        self,
        field_name: str,
        index_type: IndexType | None = None,
        sort_order: IndexSortOrder = IndexSortOrder.ASCENDING,
        expires_after_seconds: float | None = None,  # For TTL indexes
        is_sparse: bool | None = None,
        is_unique: bool | None = None,
        is_hidden: bool | None = None,
        collation_locale: str = "en",
        collation_strength: CollationStrength | None = None,
        partial_expression: CollectionFilterExpression | None = None,
        index_name: str | None = None,
    ):
        self.field_name = field_name
        self.index_type = index_type
        self.sort_order = sort_order
        self.expires_after_seconds = expires_after_seconds
        self.is_sparse = is_sparse
        self.is_unique = is_unique
        self.is_hidden = is_hidden
        self.collation_locale = collation_locale
        self.collation_strength = collation_strength
        self.partial_expression = partial_expression
        self.index_name = index_name

    def serialize(self) -> dict[str, Any]:
        """Serialize the index definition into a MongoDB-compatible format."""
        if self.index_type:
            return {self.field_name: self.index_type}
        return {self.field_name: self.sort_order}

    def build_kwargs(self) -> dict[str, Any]:
        """Additional specifications for the index to be passed to MongoDB's create_index.

        Only includes non-null values.
        """
        kwargs: dict[str, Any] = {}

        if self.index_name:
            kwargs["name"] = self.index_name

        if self.expires_after_seconds is not None:
            if self.index_type in (IndexType.TEXT, IndexType.HASHED):
                # TODO: Warning log that the TTL will not be applied as
                # it's not supported with Text
                pass
            else:
                kwargs["expireAfterSeconds"] = self.expires_after_seconds

        if self.is_sparse is not None:
            kwargs["sparse"] = self.is_sparse

        if self.is_unique is not None:
            kwargs["unique"] = self.is_unique

        if self.is_hidden is not None:
            kwargs["hidden"] = self.is_hidden

        if self.collation_strength is not None or self.collation_locale:
            kwargs["collation"] = {
                "locale": self.collation_locale,
                **({"strength": self.collation_strength.value} if self.collation_strength is not None else {}),
            }

        if self.partial_expression is not None:
            kwargs["partialFilterExpression"] = self.partial_expression.serialize()

        return kwargs

    def __hash__(self) -> int:
        """Generate a hash based on the index properties.

        Returns:
            int: A hash value representing the index.
        """
        return hash(
            (
                self.field_name,
                self.index_type,
                self.sort_order,
                self.expires_after_seconds,
                self.is_sparse,
                self.is_unique,
                self.is_hidden,
                self.collation_locale,
                self.collation_strength,
                self.index_name,
                self.partial_expression,
            )
        )

    def __eq__(self, expr: object) -> bool:
        """Compare this index expression with another for equality.

        Args:
            expr (object): The expression to compare against.

        Returns:
            bool: True if the expressions are equal, False otherwise.
        """
        if isinstance(expr, self.__class__):
            return self.build_kwargs() == expr.build_kwargs() and self.serialize() == expr.serialize()

        if isinstance(expr, str):
            return self.field_name == expr
        return False


class IndexExpressionBuilder:
    """Builder pattern for constructing IndexExpression instances with a fluent interface."""

    def __init__(self, field_name: str):
        self.field_name = field_name
        self.kwargs: dict[str, Any] = {"field_name": field_name}

    def use_sort_order(self, sort_order: IndexSortOrder) -> "IndexExpressionBuilder":
        """Set the sort order for the index.

        Args:
            sort_order (IndexSortOrder): The sort order to use.

        Returns:
            IndexExpressionBuilder: The builder instance for chaining.
        """
        self.kwargs["sort_order"] = sort_order
        return self

    def use_index_type(self, index_type: IndexType) -> "IndexExpressionBuilder":
        """Set the index type for the index.

        Args:
            index_type (IndexType): The index type to use.

        Returns:
            IndexExpressionBuilder: The builder instance for chaining.
        """
        self.kwargs["index_type"] = index_type
        return self

    def use_collation(self, locale: str, strength: CollationStrength) -> "IndexExpressionBuilder":
        """Set the collation settings for the index.

        Args:
            locale (str): The locale for string comparison.
            strength (CollationStrength): The collation strength to use.

        Returns:
            IndexExpressionBuilder: The builder instance for chaining.
        """
        self.kwargs["collation_locale"] = locale
        self.kwargs["collation_strength"] = strength
        return self

    def use_partial_expression(self, expression: CollectionFilterExpression) -> "IndexExpressionBuilder":
        """Set a partial filter expression for the index.

        Args:
            expression (CollectionFilterExpression): The partial filter expression to use.

        Returns:
            IndexExpressionBuilder: The builder instance for chaining.
        """
        self.kwargs["partial_expression"] = expression
        return self

    def use_index_name(self, index_name: str) -> "IndexExpressionBuilder":
        """Set a custom name for the index.

        Args:
            index_name (str): The custom name for the index.

        Returns:
            IndexExpressionBuilder: The builder instance for chaining.
        """
        self.kwargs["index_name"] = index_name
        return self

    def use_sparse(self, is_sparse: bool = True) -> "IndexExpressionBuilder":
        """Set whether the index should be sparse.

        Args:
            is_sparse (bool): Whether the index should be sparse. Defaults to True.

        Returns:
            IndexExpressionBuilder: The builder instance for chaining.
        """
        self.kwargs["is_sparse"] = is_sparse
        return self

    def use_unique(self, is_unique: bool = True) -> "IndexExpressionBuilder":
        """Set whether the index should enforce uniqueness.

        Args:
            is_unique (bool): Whether the index should enforce uniqueness. Defaults to True.

        Returns:
            IndexExpressionBuilder: The builder instance for chaining.
        """
        self.kwargs["is_unique"] = is_unique
        return self

    def use_hidden(self, is_hidden: bool = True) -> "IndexExpressionBuilder":
        """Set whether the index should be hidden.

        Args:
            is_hidden (bool): Whether the index should be hidden. Defaults to True.

        Returns:
            IndexExpressionBuilder: The builder instance for chaining.
        """
        self.kwargs["is_hidden"] = is_hidden
        return self

    def use_ttl(self, seconds: float) -> "IndexExpressionBuilder":
        """Set the TTL (time-to-live) for the index.

        Args:
            seconds (float): The time-to-live value in seconds.

        Returns:
            IndexExpressionBuilder: The builder instance for chaining.
        """
        self.kwargs["expires_after_seconds"] = seconds
        return self

    def build_index(self) -> IndexExpression:
        """Construct the IndexExpression instance with the configured properties.

        Returns:
            IndexExpression: The constructed IndexExpression instance.
        """
        return IndexExpression(**self.kwargs)



from collections.abc import Iterable
from typing import Any

from pymongo import MongoClient
from pymongo.errors import PyMongoError

from pydongo.drivers.base import AbstractSyncMongoDBDriver
from pydongo.expressions.index import IndexExpression


class PyMongoDriver(AbstractSyncMongoDBDriver):
    """Default synchronous MongoDB driver implementation using PyMongo.

    This driver connects to a real MongoDB instance and executes
    blocking operations such as insert, update, query, and delete.
    """

    def __init__(self, connection_string: str, database_name: str):
        """Initialize the driver with a MongoDB connection URI and database name.

        Args:
            connection_string (str): MongoDB URI (e.g., "mongodb://localhost:27017").
            database_name (str): The target database to operate on.
        """
        self.connection_string = connection_string
        self.database_name = database_name
        self.client: MongoClient | None = None  # type: ignore[type-arg]

    def connect(self) -> bool:
        """Establish a connection to the MongoDB server.

        Returns:
            bool: True if the connection is successful.

        Raises:
            RuntimeError: If connection to MongoDB fails.
        """
        try:
            self.client = MongoClient(self.connection_string)
            self.db = self.client[self.database_name]
            return True
        except PyMongoError as e:
            raise RuntimeError(f"Failed to connect to MongoDB: {e}") from e

    def close(self) -> None:
        """Close the MongoDB connection."""
        if self.client:
            self.client.close()

    def insert_one(self, collection: str, document: dict[str, Any]) -> dict[str, Any]:
        """Insert a single document into a collection.

        Args:
            collection (str): Name of the collection.
            document (dict): Document to insert.

        Returns:
            dict: Result of the insert operation, including the inserted ID.
        """
        result = self.db[collection].insert_one(document)
        return {"inserted_id": str(result.inserted_id)}

    def insert_many(self, collection: str, documents: list[dict[str, Any]]) -> dict[str, Any]:
        """Insert multiple documents into a collection.

        Args:
            collection (str): Name of the collection.
            documents (list): List of documents to insert.

        Returns:
            dict: Result including inserted IDs.
        """
        result = self.db[collection].insert_many(documents)
        return {"inserted_ids": [str(_id) for _id in result.inserted_ids]}

    def find_one(self, collection: str, query: dict[str, Any]) -> dict[str, Any] | None:
        """Find a single document that matches the query.

        Args:
            collection (str): Name of the collection.
            query (dict): MongoDB filter query.

        Returns:
            dict or None: The found document, or None if not found.
        """
        return self.db[collection].find_one(query)

    def find_many(
        self,
        collection: str,
        query: dict[str, Any],
        sort_criteria: dict[str, Any],
        offset: int | None = None,
        limit: int | None = None,
    ) -> Iterable[dict[str, Any]]:
        """Find multiple documents that match the query.

        Args:
            collection (str): Name of the collection.
            query (dict): Filter conditions.
            sort_criteria (dict): Sorting fields and directions.
            offset (Union[int, None]): Number of records to skip.
            limit (Union[int, None]): Maximum number of documents to return.

        Returns:
            Iterable[dict]: Cursor for matching documents.
        """
        cursor = self.db[collection].find(query)

        if sort_criteria:
            cursor = cursor.sort(list(sort_criteria.items()))

        if offset:
            cursor = cursor.skip(offset)

        if limit:
            cursor = cursor.limit(limit)

        return cursor

    def update_one(
        self,
        collection: str,
        query: dict[str, Any],
        update: dict[str, Any],
        upsert: bool = False,
    ) -> dict[str, Any]:
        """Update a single document matching the query.

        Args:
            collection (str): Name of the collection.
            query (dict): Filter conditions.
            update (dict): Update operations (e.g., {"$set": {...}}).
            upsert (bool): Whether to insert the document if it doesn't exist.

        Returns:
            dict: Update result with match, modification, and upsert info.
        """
        result = self.db[collection].update_one(query, update, upsert=upsert)
        return {
            "matched_count": result.matched_count,
            "modified_count": result.modified_count,
            "upserted_id": str(result.upserted_id) if result.upserted_id else None,
        }

    def delete_one(self, collection: str, query: dict[str, Any]) -> dict[str, Any]:
        """Delete a single document matching the query.

        Args:
            collection (str): Name of the collection.
            query (dict): Filter conditions.

        Returns:
            dict: Delete result with count of deleted documents.
        """
        result = self.db[collection].delete_one(query)
        return {"deleted_count": result.deleted_count}

    def count(self, collection: str, query: dict[str, Any]) -> int:
        """Count the number of documents matching the query.

        Args:
            collection (str): Name of the collection.
            query (dict): Filter conditions.

        Returns:
            int: Number of documents matching the filter.
        """
        return self.db[collection].count_documents(query)

    def exists(self, collection: str, query: dict[str, Any]) -> bool:
        """Check if at least one document exists that matches the query.

        Args:
            collection (str): Name of the collection.
            query (dict): Filter conditions.

        Returns:
            bool: True if a matching document exists.
        """
        return self.db[collection].count_documents(query, limit=1) > 0

    def create_index(self, collection: str, index: tuple[IndexExpression]) -> None:
        """Create an index on a collection in the MongoDB Database.

        Args:
            collection (str): The collection name.
            index (tuple[IndexExpression]):
                A tuple of IndexExpression objects representing the index to create
                NOTE: Multiple elements in tuple indicate a single compound index
                not multiple indexes

        """
        index_key: list[tuple[str, Any]] = []
        final_kwargs = {}

        for expr in index:
            index_key.extend(expr.serialize().items())
            final_kwargs.update(expr.build_kwargs())

        self.db[collection].create_index(index_key, **final_kwargs)

    def update_many(self, collection: str, query: dict[str, Any], update: dict[str, Any]) -> dict[str, Any]:
        """Update multiple documents matching the filter.

        Args:
            collection (str): Name of the collection.
            query (dict): Filter conditions.
            update (dict): Update operations.

        Returns:
            dict: Update result with count of matched and modified documents.
        """
        result = self.db[collection].update_many(query, update)
        return {
            "matched_count": result.matched_count,
            "modified_count": result.modified_count,
        }



from abc import ABC, abstractmethod
from typing import Any


class BaseExpression(ABC):
    """Base class for all MongoDB expressions."""

    @abstractmethod
    def serialize(self) -> Any:
        """Serialize an expression into a MongoDB compatible query DTO."""



from typing import Any

from pydongo.expressions.base import BaseExpression


class CollectionFilterExpression(BaseExpression):
    """Represents a boolean filter expression used to query a MongoDB collection.

    These expressions are composable using:
        - Logical OR (`|`)
        - Logical AND (`&`)
        - Logical NOT (`~`)

    Example usage:
        filter_1 = User.age > 18
        filter_2 = User.name == "Samuel"
        combined = filter_1 & filter_2

    The resulting expression can be serialized and passed to the database driver.
    """

    def __init__(self, expression: dict[str, Any] | None = None):
        """Initialize the filter expression with an optional initial dictionary.

        Args:
            expression (Union[dict, None]): A MongoDB-style query dictionary.
        """
        super().__init__()
        self.expression = expression or {}

    def with_expression(self, expression: dict[str, Any]) -> "CollectionFilterExpression":
        """Mutate the internal expression by merging in a new dictionary.

        Args:
            expression (dict): A MongoDB-style query to merge into the current expression.

        Returns:
            CollectionFilterExpression: The current instance (for chaining).
        """
        self.expression.update(expression)
        return self

    def serialize(self) -> dict[str, Any]:
        """Serialize the filter expression into a MongoDB-compatible query object.

        Returns:
            dict: Serialized MongoDB query.
        """
        return self.expression

    def __and__(self, other: "CollectionFilterExpression") -> "CollectionFilterExpression":
        """Combine this filter with another using a logical AND.

        Args:
            other (CollectionFilterExpression): Another filter to combine with.

        Returns:
            CollectionFilterExpression: A new combined filter.
        """
        expression = {"$and": [self.expression, other.expression]}
        return CollectionFilterExpression(expression=expression)

    def __or__(self, other: "CollectionFilterExpression") -> "CollectionFilterExpression":
        """Combine this filter with another using a logical OR.

        Args:
            other (CollectionFilterExpression): Another filter to combine with.

        Returns:
            CollectionFilterExpression: A new combined filter.
        """
        expression = {"$or": [self.expression, other.expression]}
        return CollectionFilterExpression(expression=expression)

    def __invert__(self) -> "CollectionFilterExpression":
        """Invert the current filter using a logical NOT.

        Returns:
            CollectionFilterExpression: A new filter expression wrapped in `$not`.
        """
        expression = {"$not": self.expression}
        return CollectionFilterExpression(expression=expression)



"""pydongo."""

import importlib.metadata

__version__ = importlib.metadata.version("pydongo")


from pydongo.drivers.async_mongo import PyMongoAsyncDriver
from pydongo.drivers.motor import MotorMongoDBDriver
from pydongo.drivers.sync_mongo import PyMongoDriver
from pydongo.workers.collection import CollectionWorker, as_collection
from pydongo.workers.document import AsyncDocumentWorker, DocumentWorker, as_document

__all__ = [
    "AsyncDocumentWorker",
    "CollectionWorker",
    "DocumentWorker",
    "MotorMongoDBDriver",
    "PyMongoAsyncDriver",
    "PyMongoDriver",
    "__version__",
    "as_collection",
    "as_document",
]



import contextlib
from abc import ABC, abstractmethod
from collections.abc import Iterable, Sequence
from typing import Any, Generic, Self, TypeVar

from pydantic import BaseModel

from pydongo.drivers.base import AbstractAsyncMongoDBDriver, AbstractMongoDBDriver, AbstractSyncMongoDBDriver
from pydongo.expressions.field import FieldExpression
from pydongo.expressions.filter import CollectionFilterExpression
from pydongo.expressions.index import IndexExpression
from pydongo.expressions.mutation import MutationExpression, MutationExpressionContext
from pydongo.utils.annotations import resolve_annotation
from pydongo.utils.serializer import restore_unserializable_fields
from pydongo.workers.document import AsyncDocumentWorker, BaseDocumentWorker, DocumentWorker

T = TypeVar("T", bound=BaseModel)


def as_collection(
    pydantic_model: type[T],
    driver: AbstractSyncMongoDBDriver | AbstractAsyncMongoDBDriver,
    collection_name: str | None = None,
) -> "CollectionWorker[T]":
    """Wraps a Pydantic model in a CollectionWorker for querying and interacting with the corresponding MongoDB collection.

    Args:
        pydantic_model (type[T]): The Pydantic model class to wrap.
        driver (Union[AbstractSyncMongoDBDriver, AbstractAsyncMongoDBDriver]): The MongoDB driver to use.
        collection_name (Union[str, None]): The name of the MongoDB collection. Defaults to None.

    Returns:
        CollectionWorker: A CollectionWorker instance.
    """  # noqa: E501
    return CollectionWorker(
        pydantic_model=pydantic_model,
        driver=driver,
        collection_name=collection_name,
    )


def _get_field_expression(name: str, model: type[T]) -> FieldExpression:
    if name not in model.model_fields:
        raise AttributeError(f"'{model.__name__}' has no field named '{name}'")
    annotation = model.model_fields[name].annotation
    annotation = resolve_annotation(annotation=annotation)

    return FieldExpression.get_field_expression(name, annotation)


class CollectionWorker(Generic[T]):
    """Entry point for querying and interacting with a MongoDB collection.

    Wraps a Pydantic model and MongoDB driver, and exposes queryable field expressions
    and high-level `find_one`, `afind_one`, and `find` methods.

    Supports both sync and async drivers via conditional branching.
    """

    def __init__(
        self,
        pydantic_model: type[T],
        driver: AbstractSyncMongoDBDriver | AbstractAsyncMongoDBDriver,
        collection_name: str | None = None,
    ):
        self.pydantic_model = pydantic_model
        self.driver = driver
        self._indexes: set[tuple[IndexExpression]] = set()
        self.__collection_name = collection_name

        self.response_builder_class = (
            AsyncCollectionResponseBuilder
            if issubclass(type(driver), AbstractAsyncMongoDBDriver)
            else SyncCollectionResponseBuilder
        )

    def find_one(self, expression: CollectionFilterExpression) -> BaseDocumentWorker[T] | None:
        """Query the database for a single document that matches the filter expression.

        Only works with a synchronous driver.

        Args:
            expression (CollectionFilterExpression): The MongoDB-style filter expression.

        Returns:
            Union[DocumentWorker, None]: A wrapped Pydantic object if found, otherwise None.
        """
        if issubclass(type(self.driver), AbstractAsyncMongoDBDriver):
            raise AttributeError("Use the afind_one method instead as you're using an async driver")
        result = self.driver.find_one(self.collection_name, expression.serialize())

        return (
            CollectionResponseBuilder.serialize_document(
                document=result,  # type: ignore[arg-type]
                document_worker_class=DocumentWorker,
                pydantic_model=self.pydantic_model,
                driver=self.driver,
            )
            if result
            else None
        )

    async def afind_one(self, expression: CollectionFilterExpression) -> BaseDocumentWorker[T] | None:
        """Asynchronously query the database for a single document matching the filter expression.

        Only works with an async driver.

        Args:
            expression (CollectionFilterExpression): The MongoDB-style filter expression.

        Returns:
            Union[BaseDocumentWorker[T], None]: A wrapped Pydantic object if found, otherwise None.
        """
        if issubclass(type(self.driver), AbstractSyncMongoDBDriver):
            raise AttributeError("Use the find_one method instead as you're using a sync driver")
        result = await self.driver.find_one(self.collection_name, expression.serialize())

        return (
            CollectionResponseBuilder.serialize_document(
                document=result,
                document_worker_class=AsyncDocumentWorker,
                pydantic_model=self.pydantic_model,
                driver=self.driver,
            )
            if result
            else None
        )

    def find(self, expression: CollectionFilterExpression | None = None) -> "CollectionResponseBuilder[T]":
        """Initiates a fluent response builder chain for querying multiple documents.

        Args:
            expression (Union[CollectionFilterExpression, None]): Optional filter expression.

        Returns:
            CollectionResponseBuilder: Builder for fluent chaining (e.g., .limit(), .sort()).
        """
        expression = expression or CollectionFilterExpression({})

        return self.response_builder_class(
            expression=expression,
            pydantic_model=self.pydantic_model,
            driver=self.driver,  # type: ignore[arg-type]
            collection_name=self.collection_name,
            indexes=self._indexes,
        )

    def use_index(
        self,
        index: IndexExpression | tuple[IndexExpression, ...],
    ) -> Self:
        """Registers an index (or compound index) on the collection.

        Accepts IndexExpression(s)

        Args:
            index: A single IndexExpression, or a tuple of them.

        Returns:
            CollectionWorker: Self, for method chaining.
        """
        index = (index,) if isinstance(index, IndexExpression) else index
        self._indexes.add(index)  # type: ignore[arg-type]
        return self

    def __getattr__(self, name: str) -> FieldExpression:
        """Returns a field expression object to enable DSL-style querying using comparison operators or dot notation for nested objects.

        Args:
            name (str): Field name in the Pydantic model.

        Returns:
            FieldExpression: Expression object tied to the field.

        Raises:
            AttributeError: If the field does not exist on the model.
        """  # noqa: E501
        return _get_field_expression(name, self.pydantic_model)

    @property
    def collection_name(self) -> str:
        """Derives the MongoDB collection name from the model or falls back to the class name.

        Returns:
            str: The name of the collection.
        """
        if self.__collection_name is None:
            self.__collection_name = str(
                self.pydantic_model.model_config.get(
                    "collection_name",
                    self.pydantic_model.__name__.lower().rstrip("s") + "s",
                )
            )

        return self.__collection_name


class CollectionResponseBuilder(ABC, Generic[T]):
    """Fluent builder for composing queries and retrieving lists of documents.

    Can build the final MongoDB query structure with sort, skip, and limit options.
    """

    def __init__(
        self,
        expression: CollectionFilterExpression,
        pydantic_model: type[T],
        collection_name: str,
        indexes: Iterable[tuple[IndexExpression]],
        **kwargs: dict[str, Any],
    ):
        self.model = pydantic_model
        self._expression = expression
        self._sort_criteria: Sequence[FieldExpression] = []
        self._limit: int | None = None
        self._offset: int | None = None

        self._indexes = indexes
        self._indexes_created = False

        self.collection_name = collection_name

        self.__mutation_ctx = MutationExpressionContext()
        self.__mutation_ctx.__enter__()

        self._other_kwargs = kwargs

    def skip(self, offset: int) -> "CollectionResponseBuilder[T]":
        """Skip the first N documents in the query result.

        Args:
            offset (int): Number of documents to skip.

        Returns:
            CollectionResponseBuilder: For chaining.
        """
        self._offset = offset
        return self

    def limit(self, limit_value: int) -> "CollectionResponseBuilder[T]":
        """Limit the number of documents returned.

        Args:
            limit_value (int): Maximum number of documents to return.

        Returns:
            CollectionResponseBuilder: For chaining.
        """
        self._limit = limit_value
        return self

    def sort(self, sort_criteria: FieldExpression | Sequence[FieldExpression]) -> "CollectionResponseBuilder[T]":
        """Apply sort criteria to the query.

        Args:
            sort_criteria (Union[FieldExpression, Sequence[FieldExpression]]): Sort instructions.

        Returns:
            CollectionResponseBuilder: For chaining.
        """
        self._sort_criteria = [sort_criteria] if isinstance(sort_criteria, FieldExpression) else sort_criteria
        return self

    def build_kwargs(self) -> dict[str, Any]:
        """Assembles the query, sort, offset, and limit into a single dictionary.

        Returns:
            dict: MongoDB-compatible query parameters.
        """
        query = self._expression.serialize()
        sortby = {expr.field_name: 1 if expr.sort_ascending else -1 for expr in self._sort_criteria}

        return {
            "query": query,
            "sort_criteria": sortby,
            "offset": self._offset,
            "limit": self._limit,
        }

    def get_mutations(self) -> dict[str, Any]:
        """Get current mutation context."""
        if MutationExpressionContext.has_context():
            return MutationExpressionContext.get_mutations()
        return {}

    def mutate(self) -> dict[str, Any]:
        """Run mutations through driver."""
        if isinstance(self.driver, AbstractSyncMongoDBDriver):
            update = self.driver.update_many(self.collection_name, self._expression.serialize(), self.get_mutations())

            self.__mutation_ctx.clear()
            return update

        raise AttributeError("Use the amutate method instead for async drivers")

    async def amutate(self) -> dict[str, Any]:
        """Run mutations through driver."""
        if isinstance(self.driver, AbstractAsyncMongoDBDriver):
            update = await self.driver.update_many(
                self.collection_name, self._expression.serialize(), self.get_mutations()
            )
            self.__mutation_ctx.clear()
            return update

        raise AttributeError("Use the mutate method instead for sync drivers")

    @classmethod
    def serialize_document(
        cls,
        document: dict[str, Any],
        document_worker_class: type[BaseDocumentWorker[T]],
        pydantic_model: type[T],
        driver: AbstractMongoDBDriver,
    ) -> BaseDocumentWorker[T]:
        """Deserialize a raw MongoDB document and wrap it in a DocumentWorker.

        Args:
            document (dict): Raw document from the database.
            document_worker_class (Type): Either DocumentWorker or AsyncDocumentWorker.
            pydantic_model (Type): Model class to hydrate.
            driver (AbstractMongoDBDriver): Driver used to perform operations.

        Returns:
            BaseDocumentWorker: Wrapped document instance.
        """
        deserialized_doc = restore_unserializable_fields(document=document)
        pydantic_object = pydantic_model(**deserialized_doc)
        object_id = deserialized_doc.get("_id")
        return document_worker_class(pydantic_object=pydantic_object, driver=driver, objectId=object_id)

    @property
    @abstractmethod
    def driver(self) -> AbstractSyncMongoDBDriver | AbstractAsyncMongoDBDriver:
        """The MongoDB driver instance.

        Returns:
            AbstractMongoDBDriver: The driver used for database operations.
        """

    def __setattr__(self, name: str, value: Any) -> None:
        """Push a mutation when setting an attribute on the fictional document worker.

        Args:
            name (str): The field name.
            value (Any): The value to set.

        Returns:
            None
        """
        if name == "model":
            return super().__setattr__(name, value)

        if name not in self.model.model_fields:
            return super().__setattr__(name, value)

        if self.__mutation_ctx.has_context():
            if isinstance(value, MutationExpression):
                MutationExpressionContext.add_mutation(value)
                return None

            expression: FieldExpression = _get_field_expression(
                name=name,
                model=self.model,
            )

            MutationExpressionContext.add_mutation(expression.set_value(value))
            return None
        return None

    def __getattr__(self, name: str) -> FieldExpression:
        """Returns a field expression object to enable DSL-style querying using comparison operators or dot notation for nested objects.

        Args:
            name (str): Field name in the Pydantic model.

        Returns:
            FieldExpression: Expression object tied to the field.

        Raises:
            AttributeError: If the field does not exist on the model.
        """  # noqa: E501
        return _get_field_expression(name, self.model)

    def __del__(self) -> None:
        """Exit the mutation context."""
        if hasattr(self, "__mutation_ctx"):
            with contextlib.suppress(Exception):
                self.__mutation_ctx.__exit__(None, None, None)


class SyncCollectionResponseBuilder(CollectionResponseBuilder[T]):
    """Response builder for synchronous drivers.

    Provides direct methods to count, chec§k existence, and iterate documents.
    """

    def __init__(
        self,
        expression: CollectionFilterExpression,
        pydantic_model: type[T],
        driver: AbstractSyncMongoDBDriver,
        collection_name: str,
        indexes: Iterable[tuple[IndexExpression]],
    ):
        if issubclass(type(driver), AbstractAsyncMongoDBDriver):
            raise AttributeError("Use the AsyncCollectionResponseBuilder instead as you're using an async driver")

        super().__init__(expression, pydantic_model, collection_name, indexes)
        self.__driver = driver

    def exists(self) -> bool:
        """Check if any document matches the filter expression.

        Returns:
            bool: True if at least one document exists.
        """
        self.create_indexes()
        return self.driver.exists(self.collection_name, self._expression.serialize())

    def count(self) -> int:
        """Count how many documents match the filter expression.

        Returns:
            int: Number of matching documents.
        """
        self.create_indexes()
        return self.driver.count(self.collection_name, self._expression.serialize())

    def all(self) -> Iterable[DocumentWorker]:
        """Retrieve all documents that match the current query builder state.

        Returns:
            Iterable[DocumentWorker]: Generator of hydrated documents.
        """
        self.create_indexes()
        kwargs = self.build_kwargs()
        documents = self.driver.find_many(collection=self.collection_name, **kwargs)

        for doc in documents:
            yield self.serialize_document(
                document=doc,
                document_worker_class=DocumentWorker,
                pydantic_model=self.model,
                driver=self.driver,
            )

    def create_indexes(self) -> None:
        """Creates indexes on the MongoDB database."""
        if self._indexes_created:
            return
        for index in self._indexes:
            self.driver.create_index(self.collection_name, index)
        self._indexes_created = True

    @property
    def driver(self) -> AbstractSyncMongoDBDriver:
        """The MongoDB driver instance.

        Returns:
            AbstractSyncMongoDBDriver: The driver used for database operations.
        """
        return self.__driver


class AsyncCollectionResponseBuilder(CollectionResponseBuilder[T]):
    """Response builder for asynchronous drivers.

    Supports async versions of exists, count, and all().
    """

    def __init__(
        self,
        expression: CollectionFilterExpression,
        pydantic_model: type[T],
        driver: AbstractAsyncMongoDBDriver,
        collection_name: str,
        indexes: Iterable[tuple[IndexExpression]],
    ):
        super().__init__(expression, pydantic_model, collection_name, indexes)
        if issubclass(type(driver), AbstractSyncMongoDBDriver):
            raise AttributeError("Use the SyncCollectionResponseBuilder instead as you're using a sync driver")
        self.__driver = driver

    async def exists(self) -> bool:
        """Asynchronously check if any document matches the filter.

        Returns:
            bool: True if a match exists.
        """
        await self.create_indexes()
        return await self.driver.exists(self.collection_name, self._expression.serialize())

    async def count(self) -> int:
        """Asynchronously count how many documents match the filter.

        Returns:
            int: Number of matching documents.
        """
        await self.create_indexes()
        return await self.driver.count(self.collection_name, self._expression.serialize())

    async def all(self) -> Iterable[AsyncDocumentWorker]:
        """Asynchronously retrieve all matching documents.

        Returns:
            Iterable[AsyncDocumentWorker]: List of wrapped async document objects.
        """
        await self.create_indexes()
        kwargs = self.build_kwargs()
        documents = await self.driver.find_many(collection=self.collection_name, **kwargs)

        if hasattr(documents, "__aiter__"):
            return [
                self.serialize_document(
                    document,
                    AsyncDocumentWorker,
                    pydantic_model=self.model,
                    driver=self.driver,
                )
                async for document in documents
            ]

        return [
            self.serialize_document(
                document,
                AsyncDocumentWorker,
                pydantic_model=self.model,
                driver=self.driver,
            )
            for document in documents
        ]

    async def create_indexes(self) -> None:
        """Asynchronously creates indexes on the MongoDB database."""
        if self._indexes_created:
            return
        for index in self._indexes:
            await self.driver.create_index(self.collection_name, index)
        self._indexes_created = True

    @property
    def driver(self) -> AbstractAsyncMongoDBDriver:
        """The MongoDB driver instance.

        Returns:
            AbstractAsyncMongoDBDriver: The driver used for database operations.
        """
        return self.__driver



from collections.abc import Iterable, Sequence
from typing import Any, get_args, get_origin

from pydantic import BaseModel

from pydongo.expressions.filter import CollectionFilterExpression
from pydongo.expressions.index import IndexExpression, IndexExpressionBuilder, IndexSortOrder, IndexType
from pydongo.expressions.mutation import MutationExpression, MutationExpressionContext
from pydongo.utils.annotations import resolve_annotation
from pydongo.utils.serializer import HANDLER_MAPPING


class FieldExpression:
    """Represents a scalar field in a MongoDB query.

    This class supports operator overloading to build query expressions using:
        ==, !=, >, >=, <, <=

    It also supports nested attribute access (e.g., user.address.city).
    """

    def __init__(
        self, field_name: str, annotation: Any = None, sort_ascending: bool = True, update_alias: str | None = None
    ) -> None:
        """Initialize a field expression.

        Args:
            field_name (str): The full dot-path name of the field.
            annotation (Any): The Pydantic type annotation for the field.
            sort_ascending (bool): Whether sorting on this field is ascending by default.
            update_alias (str): Custom alias for update expressions
        """
        self.annotation = annotation
        self._field_name = field_name
        self._update_alias = update_alias or field_name
        self.sort_ascending = sort_ascending

    def to_index(self) -> IndexExpression:
        """Returns a complete, ready-to-use `IndexExpression` for this field.

        This is a convenience method for quickly generating a basic index configuration
        using the default sort order and inferred index type (e.g., TEXT for strings).
        """
        return self.as_index().build_index()

    def as_index(self) -> IndexExpressionBuilder:
        """Returns an `IndexExpressionBuilder` initialized with this field.

        This allows users to customize the index further, such as adding uniqueness,
        TTL, collation, or partial filter expressions—before building the final index object.
        """
        sort_order = IndexSortOrder.ASCENDING if self.sort_ascending else IndexSortOrder.DESCENDING

        builder = IndexExpressionBuilder(field_name=self.field_name).use_sort_order(sort_order=sort_order)

        datatype = get_origin(self.annotation) or self.annotation

        if issubclass(datatype, str):
            return builder.use_index_type(IndexType.TEXT)

        return builder

    def unset(self) -> MutationExpression:
        """Build an unset expression.

        Returns:
            MutationExpression: A MongoDB `$unset` expression.
        """
        mutation = MutationExpression(field_name=self._update_alias, operation="$unset", value="")
        self.add_mutation(mutation)
        return mutation

    def set_value(self, value: Any) -> MutationExpression:
        """Build a set expression.

        Args:
            value (Any): The value to set.

        Returns:
            MutationExpression: A MongoDB `$set` expression.
        """
        return MutationExpression(field_name=self._update_alias, operation="$set", value=value)

    def _get_comparative_expression(self, operator: str, value: Any) -> dict[str, Any]:
        """Build a MongoDB filter expression using the given operator and value.

        Args:
            operator (str): MongoDB comparison operator (e.g., "$gt").
            value (Any): The value to compare against.

        Returns:
            dict: A MongoDB-compatible filter clause.
        """
        value_class = type(value)
        if value_class in HANDLER_MAPPING:
            value = HANDLER_MAPPING[value_class].serialize(value)

        return {self.field_name: {operator: value}}

    def __eq__(self, other: object) -> "CollectionFilterExpression":  # type: ignore  # noqa: PGH003
        """Build an equality expression (`==`)."""
        return CollectionFilterExpression().with_expression(self._get_comparative_expression("$eq", other))

    def __ne__(self, other: object) -> "CollectionFilterExpression":  # type: ignore  # noqa: PGH003
        """Build an inequality expression (`!=`)."""
        return CollectionFilterExpression().with_expression(self._get_comparative_expression("$ne", other))

    def __gt__(self, other: Any) -> "CollectionFilterExpression":
        """Build a greater-than expression (`>`)."""
        return CollectionFilterExpression().with_expression(self._get_comparative_expression("$gt", other))

    def __ge__(self, other: Any) -> "CollectionFilterExpression":
        """Build a greater-than-or-equal expression (`>=`)."""
        return CollectionFilterExpression().with_expression(self._get_comparative_expression("$gte", other))

    def __lt__(self, other: Any) -> "CollectionFilterExpression":
        """Build a less-than expression (`<`)."""
        return CollectionFilterExpression().with_expression(self._get_comparative_expression("$lt", other))

    def __le__(self, other: Any) -> "CollectionFilterExpression":
        """Build a less-than-or-equal expression (`<=`)."""
        return CollectionFilterExpression().with_expression(self._get_comparative_expression("$lte", other))

    def __neg__(self) -> "FieldExpression":
        """Flip the default sort order.

        Returns:
            FieldExpression: A new instance with reversed sort order.
        """
        return FieldExpression(
            field_name=self.field_name,
            annotation=self.annotation,
            sort_ascending=not self.sort_ascending,
        )

    @property
    def field_name(self) -> str:
        """The full field name, including nested paths if applicable."""
        return self._field_name

    def __getattr__(self, name: str) -> "FieldExpression":
        """Support for chained dot notation, e.g. `user.address.city`.

        Args:
            name (str): Subfield name.

        Returns:
            FieldExpression: A new field expression for the nested field.
        """
        return self._getattr(self.field_name, self.annotation, name)

    def __setattr__(self, name: str, value: Any) -> None:
        """Append a mutation to context."""
        if not MutationExpressionContext.has_context():
            return super().__setattr__(name, value)

        if name == "annotation":
            return super().__setattr__(name, value)

        annotation = resolve_annotation(annotation=self.annotation)
        return self._setattr(name, value, annotation)

    def _setattr(self, name: str, value: Any, annotation: Any) -> None:
        """Append a mutation to context."""
        if not issubclass(annotation, BaseModel):
            return super().__setattr__(name, value)
        if name not in annotation.model_fields:
            return super().__setattr__(name, value)

        field = self._getattr(self.field_name, annotation, name)

        if isinstance(value, MutationExpression):
            self.add_mutation(value)
            return None

        self.add_mutation(field.set_value(value))
        return None

    def add_mutation(self, value: MutationExpression) -> None:
        """Helper function to add a mutation."""
        if MutationExpressionContext.has_context():
            MutationExpressionContext.add_mutation(value)

    @classmethod
    def _getattr(cls, field_name: str, annotation: Any, name: str) -> "FieldExpression":
        """Helper to resolve nested fields.

        Args:
            field_name (str): Current field name path.
            annotation (Any): Type annotation of the current field.
            name (str): Next subfield name.

        Returns:
            FieldExpression: A new expression with nested field access.

        Raises:
            AttributeError: If the annotation is not a Pydantic model or field is invalid.
        """
        annotation = resolve_annotation(annotation=annotation)
        field_name = ".".join((field_name, name))

        if not issubclass(annotation, BaseModel):
            raise AttributeError(f"'{annotation.__name__}' is not an object but a scalar value")
        if name not in annotation.model_fields:
            raise AttributeError(f"'{annotation.__name__}' has no field named '{name}'")

        annotation = annotation.model_fields[name].annotation
        return cls.get_field_expression(field_name, annotation)

    @staticmethod
    def get_field_expression(field_name: str, annotation: Any) -> "FieldExpression":
        """Factory method to return the appropriate FieldExpression subclass."""
        dtype = get_origin(annotation) or annotation
        if isinstance(dtype, type) and issubclass(dtype, (Sequence, set)):
            return ArrayFieldExpression(field_name, annotation=annotation)
        if isinstance(dtype, type) and issubclass(dtype, (int, float)):
            return NumericFieldExpression(field_name, annotation=annotation)
        return FieldExpression(field_name, annotation=annotation)


class NumericFieldExpression(FieldExpression):
    """Represents a numeric field in a MongoDB document."""

    def __add__(self, other: float) -> MutationExpression:
        """Build an addition expression (`+`).

        Args:
            other (Any): The value to add.

        Returns:
            MutationExpression: A MongoDB `$expr` addition expression.
        """
        return MutationExpression(field_name=self._update_alias, operation="$inc", value=other)

    def __sub__(self, other: float) -> MutationExpression:
        """Build a subtraction expression (`-`).

        Args:
            other (Any): The value to subtract.

        Returns:
            MutationExpression: A MongoDB `$expr` subtraction expression.
        """
        return MutationExpression(field_name=self._update_alias, operation="$inc", value=-other)

    def __mul__(self, other: float) -> MutationExpression:
        """Build a multiplication expression (`*`).

        Args:
            other (Any): The value to multiply by.

        Returns:
            MutationExpression: A MongoDB `$expr` multiplication expression.
        """
        return MutationExpression(field_name=self._update_alias, operation="$mul", value=other)

    def __truediv__(self, other: float) -> MutationExpression:
        """Build a division expression (`/`).

        Args:
            other (Any): The value to divide by.

        Returns:
            MutationExpression: A MongoDB `$expr` division expression.
        """
        return MutationExpression(field_name=self._update_alias, operation="$mul", value=1 / other)

    def setmax(self, value: float) -> MutationExpression:
        """Build a set maximum expression.

        Args:
            value (Any): The value to set as maximum.

        Returns:
            MutationExpression: A MongoDB `$max` expression.
        """
        mutation = MutationExpression(field_name=self._update_alias, operation="$max", value=value)
        self.add_mutation(mutation)
        return mutation

    def setmin(self, value: float) -> MutationExpression:
        """Build a set minimum expression.

        Args:
            value (Any): The value to set as minimum.

        Returns:
            MutationExpression: A MongoDB `$min` expression.
        """
        mutation = MutationExpression(field_name=self._update_alias, operation="$min", value=value)
        self.add_mutation(mutation)
        return mutation


class ArraySizeFieldExpression(FieldExpression):
    """Represents an expression for the size of an array field.

    Used for queries like: `len(User.tags) > 2`
    """

    def _get_comparative_expression(self, operator: str, value: int) -> dict[str, Any]:
        """Build a MongoDB `$expr` query comparing the size of an array.

        Args:
            operator (str): Comparison operator (e.g., "$gt").
            value (int): Value to compare the array size against.

        Returns:
            dict: MongoDB `$expr` query.
        """
        return {"$expr": {operator: [{"$size": f"${self.field_name}"}, value]}}


class ArrayFieldExpression(FieldExpression):
    """Represents an array field in a MongoDB document.

    Adds support for array-specific operations like `.contains()`, `.size()`, and `.matches()`.
    """

    def size(self) -> ArraySizeFieldExpression:
        """Get an expression that targets the array's length.

        Returns:
            ArraySizeFieldExpression: An expression targeting the array's size.
        """
        return ArraySizeFieldExpression(self.field_name, self.annotation)

    def matches(self, values: Iterable[Any], match_order: bool = False) -> CollectionFilterExpression:
        """Check if the array exactly matches the provided values.

        Args:
            values (Iterable[Any]): Values to match against.
            match_order (bool): If True, requires order-sensitive match.

        Returns:
            CollectionFilterExpression: MongoDB `$all` or exact match query.
        """
        if match_order:
            return CollectionFilterExpression().with_expression({self.field_name: list(values)})

        expression = {self.field_name: {"$all": list(values)}}
        return CollectionFilterExpression().with_expression(expression)

    def contains(self, value: Any) -> CollectionFilterExpression:
        """Check if the array contains one or more values.

        Args:
            value (Any): A scalar or iterable to check presence for.

        Returns:
            CollectionFilterExpression: MongoDB `$in` expression.
        """
        expression = {self.field_name: {"$in": value}}
        return CollectionFilterExpression().with_expression(expression)

    def excludes(self, value: Any) -> CollectionFilterExpression:
        """Check if the array excludes one or more values.

        Args:
            value (Any): A scalar or iterable to check absence for.

        Returns:
            CollectionFilterExpression: MongoDB `$nin` expression.
        """
        expression = {self.field_name: {"$nin": value}}
        return CollectionFilterExpression().with_expression(expression)

    def push(self, value: Any) -> MutationExpression:
        """Build a push expression to add an element to the array.

        Args:
            value (Any): The value to push into the array.

        Returns:
            MutationExpression: A MongoDB `$push` expression.
        """
        mutation = MutationExpression(field_name=self.field_name, operation="$push", value=value)
        self.add_mutation(mutation)
        return mutation

    def add_to_set(self, value: Any) -> MutationExpression:
        """Build an addToSet expression to add an element to the array if it doesn't already exist.

        Args:
            value (Any): The value to add to the set.

        Returns:
            MutationExpression: A MongoDB `$addToSet` expression.
        """
        mutation = MutationExpression(field_name=self.field_name, operation="$addToSet", value=value)
        self.add_mutation(mutation)
        return mutation

    def pull(self, value: Any) -> MutationExpression:
        """Build a pull expression to remove an element from the array.

        Args:
            value (Any): The value to remove from the array.

        Returns:
            MutationExpression: A MongoDB `$pull` expression.
        """
        mutation = MutationExpression(field_name=self.field_name, operation="$pull", value=value)
        self.add_mutation(mutation)
        return mutation

    def popleft(self) -> MutationExpression:
        """Build a pop expression to remove the first element from the array.

        Returns:
            MutationExpression: A MongoDB `$pop` expression.
        """
        position = 1

        mutation = MutationExpression(field_name=self.field_name, operation="$pop", value=position)
        self.add_mutation(mutation)
        return mutation

    def popright(self) -> MutationExpression:
        """Build a pop expression to remove the last element from the array.

        Returns:
            MutationExpression: A MongoDB `$pop` expression.
        """
        position = -1

        mutation = MutationExpression(field_name=self.field_name, operation="$pop", value=position)
        self.add_mutation(mutation)
        return mutation

    def __getattr__(self, name: str) -> FieldExpression:
        """Support accessing subfields within an array of objects.

        Args:
            name (str): Name of the subfield.

        Returns:
            FieldExpression: Field expression for the nested field.

        Raises:
            TypeError: If element type can't be inferred.
        """
        args = get_args(self.annotation)
        if not args:
            raise TypeError(f"Cannot extract element type from {self.annotation}")

        element_type = args[0]
        return self._getattr(self.field_name, element_type, name)

    def __len__(self) -> ArraySizeFieldExpression:
        """Return an expression for the array's size (using `len()`).

        Returns:
            ArraySizeFieldExpression: Expression targeting array length.
        """
        return self.size()

    def __contains__(self, value: Any) -> CollectionFilterExpression:
        """Python-style containment check using `in` syntax.

        Args:
            value (Any): Element to check for.

        Returns:
            CollectionFilterExpression: MongoDB `$in` expression.
        """
        return self.contains(value=value)



from typing import Annotated, Any, Optional, Union, get_args, get_origin


def resolve_annotation(annotation: Any) -> Any:
    """Helper function to resolve actual data types from Optional, Union, or Annotated wrappers.

    Args:
        annotation (Any): Type annotation from the Pydantic model.

    Returns:
        Any: The resolved base type annotation.
    """
    origin = get_origin(annotation)
    if origin is Union or origin is Optional:
        args = get_args(annotation)
        non_none_args = [arg for arg in args if arg is not type(None)]
        return resolve_annotation(non_none_args[0] if non_none_args else annotation)
    if origin is Annotated:
        annotation = resolve_annotation(get_args(annotation)[0])

    return annotation


# def resolve_annotation(annotation: Any) -> Any:
#     """Resolve the base type from Optional, Union, Annotated, and generic containers."""

#     while True:
#         origin = get_origin(annotation)

#         # Optional / Union[T, None]
#         if origin is Union:
#             args = [arg for arg in get_args(annotation) if arg is not type(None)]
#             annotation = args[0] if args else annotation
#             continue

#         # Annotated[T, ...]
#         if origin is Annotated:
#             annotation = get_args(annotation)[0]
#             continue

#         # Built-in generics: list[T], dict[K, V], set[T], tuple[T], etc.
#         if origin is not None:
#             args = get_args(annotation)
#             if args:
#                 # Prefer the "value" type for mappings
#                 annotation = args[-1]
#                 continue

#         break

#     return annotation



from typing import Any, Generic, TypeVar, Union

from bson import ObjectId
from pydantic import BaseModel

from pydongo.drivers.base import AbstractAsyncMongoDBDriver, AbstractSyncMongoDBDriver
from pydongo.utils.serializer import replace_unserializable_fields

T = TypeVar("T", bound=BaseModel)


def as_document(
    pydantic_object: T,
    driver: AbstractSyncMongoDBDriver | AbstractAsyncMongoDBDriver,
    collection_name: str | None = None,
) -> Union["DocumentWorker", "AsyncDocumentWorker"]:
    """Wraps a Pydantic object in either a synchronous or asynchronous document worker, depending on the provided MongoDB driver.

    Args:
        pydantic_object (T): A Pydantic model instance.
        driver (Union[AbstractMongoDBDriver, AbstractAsyncMongoDBDriver]): The MongoDB driver in use.
        collection_name (Union[str, None]): Optionally set the name of the collection to persist the document to.


    Returns:
        Union[DocumentWorker, AsyncDocumentWorker]: A document wrapper with persistence methods.
    """  # noqa: E501
    if issubclass(driver.__class__, AbstractAsyncMongoDBDriver):
        return AsyncDocumentWorker(
            pydantic_object=pydantic_object,
            driver=driver,
            collection_name=collection_name,
        )
    return DocumentWorker(
        pydantic_object=pydantic_object,
        driver=driver,
        collection_name=collection_name,
    )


class BaseDocumentWorker(Generic[T]):
    """Base class for document-level operations like serialization, query generation, and collection detection.

    This class provides common functionality for both synchronous and asynchronous document workers.
    """

    def __init__(  # noqa: D417
        self,
        pydantic_object: T,
        objectId: str | None = None,  # noqa: N803
        collection_name: str | None = None,
        *args: Any,  # noqa: ARG002
        **kwargs: Any,  # noqa: ARG002
    ) -> None:
        """Initialize the document wrapper.

        Args:
            pydantic_object (T): The Pydantic model to wrap.
            objectId (Union[str, None]): MongoDB ObjectId of the document.
            collection_name (Union[str, None]): Optionally set the name of the collection to persist the document to
        """
        self.pydantic_object = pydantic_object
        self.objectId = ObjectId(objectId) if objectId else None

        self.__collection_name = collection_name

    @property
    def collection_name(self) -> str:
        """Returns the MongoDB collection name associated with the model.

        Returns:
            str: The collection name.
        """
        if self.__collection_name is None:
            self.__collection_name = str(
                self.pydantic_object.model_config.get(
                    "collection_name",
                    self.pydantic_object.__class__.__name__.lower().rstrip("s") + "s",
                )
            )

        return self.__collection_name

    def get_query(self) -> dict[str, Any]:
        """Builds a MongoDB query to uniquely identify the current document.

        Returns:
            dict: A MongoDB query dict.
        """
        if self.objectId:
            return {"_id": ObjectId(self.objectId)}
        return self.serialize()

    def serialize(self) -> dict[str, Any]:
        """Serializes the Pydantic model to a dictionary suitable for MongoDB storage.

        Returns:
            dict: Serialized representation of the document.
        """
        dump = self.pydantic_object.model_dump()
        return replace_unserializable_fields(dump)

    def __getattr__(self, name: str) -> Any:
        """Proxy attribute access to the underlying Pydantic model.

        Args:
            name (str): The attribute name.

        Returns:
            Any: Attribute value from the Pydantic model.
        """
        return getattr(self.pydantic_object, name)

    def __repr__(self) -> str:
        """Returns a human-readable string representation of the document.

        Returns:
            str: Representation of the document wrapper.
        """
        return f"{self.__class__.__name__}(<{self.pydantic_object}>)"

    def __setattr__(self, name: str, value: Any) -> None:
        """Pipe updates to the pydantic object where applicable."""
        if name == "pydantic_object":
            return super().__setattr__(name, value)

        if hasattr(self.pydantic_object, name):
            setattr(self.pydantic_object, name, value)

        return super().__setattr__(name, value)


class DocumentWorker(BaseDocumentWorker):  # type: ignore[type-arg]
    """Document wrapper for synchronous MongoDB operations.

    Provides methods to save and delete documents from MongoDB.
    """

    def __init__(
        self,
        pydantic_object: T,
        driver: AbstractSyncMongoDBDriver,
        objectId: str | None = None,  # noqa: N803
        collection_name: str | None = None,
    ) -> None:
        """Initialize the sync document worker.

        Args:
            pydantic_object (T): The Pydantic model to wrap.
            driver (AbstractSyncMongoDBDriver): The sync MongoDB driver.
            objectId (Union[str, None]): ObjectId for the document.
            collection_name (Union[str, None]): Optionally set the name of the collection to persist the document to.
        """
        super().__init__(
            pydantic_object=pydantic_object,
            objectId=objectId,
            collection_name=collection_name,
        )
        self.driver = driver

    def save(self) -> dict[str, Any]:
        """Save the document to MongoDB. Performs an insert if no ObjectId is present, or an update otherwise.

        Returns:
            dict: Result of the MongoDB insert or update operation.
        """
        payload = self.serialize()
        if self.objectId is None:
            response = self.driver.insert_one(self.collection_name, payload)
            self.objectId = response.get("inserted_id")
        else:
            query = {"_id": ObjectId(self.objectId)}
            response = self.driver.update_one(self.collection_name, query=query, update={"$set": payload})

        return response

    def delete(self) -> dict[str, Any]:
        """Delete the document from MongoDB using its ObjectId.

        Returns:
            dict: Result of the delete operation.
        """
        query = self.get_query()
        return self.driver.delete_one(collection=self.collection_name, query=query)


class AsyncDocumentWorker(BaseDocumentWorker):  # type: ignore[type-arg]
    """Document wrapper for asynchronous MongoDB operations.

    Provides async methods to save and delete documents.
    """

    def __init__(
        self,
        pydantic_object: T,
        driver: AbstractAsyncMongoDBDriver,
        objectId: str | None = None,  # noqa: N803
        collection_name: str | None = None,
    ):
        """Initialize the async document worker.

        Args:
            pydantic_object (T): The Pydantic model to wrap.
            driver (AbstractAsyncMongoDBDriver): The async MongoDB driver.
            objectId (Union[str, None]): ObjectId for the document.
            collection_name (Union[str, None]): Optionally set the name of the collection to persist the document to.
        """
        super().__init__(
            pydantic_object=pydantic_object,
            objectId=objectId,
            collection_name=collection_name,
        )
        self.driver = driver

    async def save(self) -> dict[str, Any]:
        """Save the document to MongoDB asynchronously. Inserts if no ObjectId is present, or updates otherwise.

        Returns:
            dict: Result of the insert or update operation.
        """
        payload = self.serialize()

        if self.objectId is None:
            response = await self.driver.insert_one(self.collection_name, payload)
            self.objectId = response.get("inserted_id")
        else:
            query = {"_id": ObjectId(self.objectId)}
            response = await self.driver.update_one(self.collection_name, query=query, update={"$set": payload})

        return response

    async def delete(self) -> dict[str, Any]:
        """Delete the document from MongoDB asynchronously using its ObjectId.

        Returns:
            dict: Result of the delete operation.
        """
        query = self.get_query()
        return await self.driver.delete_one(collection=self.collection_name, query=query)



import datetime
from abc import ABC, abstractmethod
from collections.abc import Mapping
from typing import Any, TypeVar
from uuid import UUID

from pydantic import BaseModel

T = TypeVar("T", bound=BaseModel)


class BaseTypeSerializer(ABC):
    """Abstract base class for type serializers.

    Implementations define how to convert values to and from MongoDB-compatible types.
    """

    @staticmethod
    @abstractmethod
    def serialize(value: Any) -> Any:
        """Convert a Python value into a MongoDB-storable format.

        Args:
            value (Any): The value to serialize.

        Returns:
            Any: A serialized version of the value.
        """

    @staticmethod
    @abstractmethod
    def deserialize(value: Any) -> Any:
        """Convert a MongoDB-stored value back to a Python-native type.

        Args:
            value (Any): The value to deserialize.

        Returns:
            Any: The original Python representation.
        """


class DateSerializer(BaseTypeSerializer):
    """Serializer for converting `datetime.date` to `datetime.datetime`.

    Ensures compatibility with MongoDB, which only supports datetime objects.
    """

    @staticmethod
    def serialize(value: datetime.date) -> datetime.datetime:
        """Serialize a `date` object into a UTC `datetime` object.

        Args:
            value (datetime.date): The date to serialize.

        Returns:
            datetime.datetime: A datetime object representing midnight of the same date.
        """
        return datetime.datetime.combine(value, datetime.datetime.min.time())

    @staticmethod
    def deserialize(value: datetime.datetime) -> datetime.date:
        """Deserialize a `datetime` object back into a `date`.

        Args:
            value (datetime.datetime): The datetime to convert.

        Returns:
            datetime.date: The date component of the datetime.
        """
        return value.date()


class UUIDSerializer(BaseTypeSerializer):
    """Serializer for UUIDs, converting them to and from strings."""

    @staticmethod
    def serialize(value: UUID) -> str:
        """Convert a UUID to a string for MongoDB storage.

        Args:
            value (UUID): The UUID to serialize.

        Returns:
            str: String representation of the UUID.
        """
        return str(value)

    @staticmethod
    def deserialize(value: str) -> UUID:
        """Convert a string back into a UUID object.

        Args:
            value (str): UUID string.

        Returns:
            UUID: The corresponding UUID object.
        """
        return UUID(value)


HANDLER_MAPPING: Mapping[Any, BaseTypeSerializer] = {
    datetime.date: DateSerializer(),
    UUID: UUIDSerializer(),
}


def replace_unserializable_fields(document: dict[str, Any]) -> dict[str, Any]:
    """Recursively replaces values in a document that are not MongoDB-compatible.

    With serialized equivalents, using registered type handlers.

    Args:
        document (dict): The original document to sanitize.

    Returns:
        dict: A deep-copied document with serializable values.
    """

    def serialize(val: Any) -> Any:
        value_type = type(val)
        if value_type in HANDLER_MAPPING:
            val = HANDLER_MAPPING[value_type].serialize(val)
        return val

    for key, value in document.items():
        if isinstance(value, dict):
            document[key] = replace_unserializable_fields(value)
        elif isinstance(value, (list, tuple, set)):
            document[key] = [serialize(element) for element in value]
        else:
            document[key] = serialize(value)

    return document


def restore_unserializable_fields(document: dict[str, Any]) -> dict[str, Any]:
    """Recursively attempts to deserialize known types in a document.

    NOTE: This is a basic placeholder strategy — it applies all registered deserializers
    without knowing the intended field type. This can result in incorrect types.

    It's recommended to use field-aware deserialization (tied to model schemas)
    for production use.

    Args:
        document (dict): A document retrieved from MongoDB.

    Returns:
        dict: A deep-copied document with deserialized values (best-effort).
    """

    def deserialize(val: Any) -> Any:
        value_type = type(val)
        if value_type in HANDLER_MAPPING:
            val = HANDLER_MAPPING[value_type].deserialize(val)
        return val

    for key, value in document.items():
        if isinstance(value, dict):
            document[key] = restore_unserializable_fields(value)
        elif isinstance(value, (list, tuple, set)):
            document[key] = [deserialize(element) for element in value]
        else:
            document[key] = deserialize(value)

    return document



import contextvars
from abc import ABC, abstractmethod
from collections.abc import Iterable
from types import TracebackType
from typing import Any, Self

from pydongo.expressions.index import IndexExpression


class AbstractMongoDBDriver(ABC):  # noqa: B024
    """Abstract base class for MongoDB drivers."""


class AbstractSyncMongoDBDriver(AbstractMongoDBDriver):
    """Abstract base class for synchronous MongoDB drivers.

    Provides context management support and coroutine/thread-safe access
    to the currently active driver instance via contextvars.
    """

    _current: contextvars.ContextVar["AbstractSyncMongoDBDriver"] = contextvars.ContextVar("mongo_driver_current")

    def __enter__(self) -> Self:
        """Enter the context manager, connect to MongoDB, and set the current driver context."""
        self.connect()
        self._token = self._current.set(self)
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit the context manager, close the connection, and reset the driver context."""
        self.close()
        self._current.reset(self._token)

    @classmethod
    def current(cls) -> "AbstractSyncMongoDBDriver":
        """Get the current MongoDB driver instance for this coroutine/thread.

        Returns:
            AbstractMongoDBDriver: The active driver instance.

        Raises:
            RuntimeError: If no driver is currently in context.
        """
        try:
            return cls._current.get()
        except LookupError as e:
            raise RuntimeError("No active MongoDBDriver context found") from e

    @abstractmethod
    def connect(self) -> bool:
        """Establish a connection to the MongoDB server.

        Returns:
            bool: True if connection was successful.
        """

    @abstractmethod
    def close(self) -> None:
        """Close the connection to the MongoDB server."""

    @abstractmethod
    def insert_one(self, collection: str, document: dict[str, Any]) -> dict[str, Any]:
        """Insert a single document into the specified collection.

        Args:
            collection (str): The collection name.
            document (dict): The document to insert.

        Returns:
            dict: Result of the insert operation.
        """

    @abstractmethod
    def insert_many(self, collection: str, documents: list[dict[str, Any]]) -> dict[str, Any]:
        """Insert multiple documents into the specified collection.

        Args:
            collection (str): The collection name.
            documents (list): List of documents to insert.

        Returns:
            dict: Result of the insert operation.
        """

    @abstractmethod
    def find_one(self, collection: str, query: dict[str, Any]) -> dict[str, Any] | None:
        """Find a single document matching the query.

        Args:
            collection (str): The collection name.
            query (dict): The filter query.

        Returns:
            dict or None: The found document or None if not found.
        """

    @abstractmethod
    def find_many(
        self,
        collection: str,
        query: dict[str, Any],
        sort_criteria: dict[str, Any],
        offset: int | None = None,
        limit: int | None = None,
    ) -> Iterable[dict[str, Any]]:
        """Find multiple documents matching the query.

        Args:
            collection (str): The collection name.
            query (dict): The filter query.
            sort_criteria (dict): How to order the results
            offset (Union[int, None]): Number of records to skip (useful for pagination)
            limit (Union[int, None]): Max number of documents to return.

        Returns:
            iterator for result: Iterable sequence of matching documents.
        """

    @abstractmethod
    def update_one(
        self,
        collection: str,
        query: dict[str, Any],
        update: dict[str, Any],
        upsert: bool = False,
    ) -> dict[str, Any]:
        """Update a single document matching the query.

        Args:
            collection (str): The collection name.
            query (dict): The filter query.
            update (dict): The update document.
            upsert (bool): Whether to insert a new document if no match is found.

        Returns:
            dict: Result of the update operation.
        """

    @abstractmethod
    def delete_one(self, collection: str, query: dict[str, Any]) -> dict[str, Any]:
        """Delete a single document matching the query.

        Args:
            collection (str): The collection name.
            query (dict): The filter query.

        Returns:
            dict: Result of the delete operation.
        """

    @abstractmethod
    def count(self, collection: str, query: dict[str, Any]) -> int:
        """Count how many records are in a collection that match the specified query.

        Args:
            collection (str): The collection name.
            query (dict): The filter query.

        Returns:
            int: Number of documents that match the specified query
        """

    @abstractmethod
    def exists(self, collection: str, query: dict[str, Any]) -> bool:
        """Check if at least one document exists in the collection that matches the specified query.

        Args:
            collection (str): The collection name.
            query (dict): The filter query.

        Returns:
            bool: True if a document exists and False if it doesn't
        """

    @abstractmethod
    def create_index(self, collection: str, index: tuple[IndexExpression]) -> None:
        """Create an index on a collection in the MongoDB Database.

        Args:
            collection (str): The collection name.
            index (tuple[IndexExpression]):
                A tuple of IndexExpression objects representing the index to create
                NOTE: Multiple elements in tuple indicate a single compound index
        """

    @abstractmethod
    def update_many(
        self,
        collection: str,
        query: dict[str, Any],
        update: dict[str, Any],
    ) -> dict[str, Any]:
        """Update multiple documents matching the query.

        Args:
            collection (str): Name of the collection.
            query (dict): MongoDB filter.
            update (dict): MongoDB update expression.

        Returns:
            dict: Result of the update operation.
        """


class AbstractAsyncMongoDBDriver(AbstractMongoDBDriver):
    """Abstract base class for asynchronous MongoDB drivers.

    Provides async context management and coroutine-safe access
    to the currently active driver instance via contextvars.
    """

    _current: contextvars.ContextVar["AbstractAsyncMongoDBDriver"] = contextvars.ContextVar("mongo_driver_current")

    async def __aenter__(self) -> Self:
        """Enter the async context manager, connect to MongoDB, and set the current driver context."""
        await self.connect()
        self._token = self._current.set(self)
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit the async context manager, close the connection, and reset the driver context."""
        await self.close()
        self._current.reset(self._token)

    @classmethod
    def current(cls) -> "AbstractAsyncMongoDBDriver":
        """Get the current MongoDB driver instance for this coroutine.

        Returns:
            AbstractAsyncMongoDBDriver: The active driver instance.

        Raises:
            RuntimeError: If no driver is currently in context.
        """
        try:
            return cls._current.get()
        except LookupError as e:
            raise RuntimeError("No active MongoDBDriver context found") from e

    @abstractmethod
    async def connect(self) -> bool:
        """Establish a connection to the MongoDB server.

        Returns:
            bool: True if connection was successful.
        """

    @abstractmethod
    async def close(self) -> None:
        """Close the connection to the MongoDB server."""

    @abstractmethod
    async def insert_one(self, collection: str, document: dict[str, Any]) -> dict[str, Any]:
        """Insert a single document into the specified collection.

        Args:
            collection (str): The collection name.
            document (dict): The document to insert.

        Returns:
            dict: Result of the insert operation.
        """

    @abstractmethod
    async def insert_many(self, collection: str, documents: list[dict[str, Any]]) -> dict[str, Any]:
        """Insert multiple documents into the specified collection.

        Args:
            collection (str): The collection name.
            documents (list): List of documents to insert.

        Returns:
            dict: Result of the insert operation.
        """

    @abstractmethod
    async def find_one(self, collection: str, query: dict[str, Any]) -> dict[str, Any] | None:
        """Find a single document matching the query.

        Args:
            collection (str): The collection name.
            query (dict): The filter query.

        Returns:
            dict or None: The found document or None if not found.
        """

    @abstractmethod
    async def find_many(
        self,
        collection: str,
        query: dict[str, Any],
        sort_criteria: dict[str, Any],
        offset: int | None = None,
        limit: int | None = None,
    ) -> Iterable[dict[str, Any]]:
        """Find multiple documents matching the query.

        Args:
            collection (str): The collection name.
            query (dict): The filter query.
            sort_criteria (dict): How to order the results
            offset (Union[int, None]): Number of records to skip (useful for pagination)
            limit (Union[int, None]): Max number of documents to return.

        Returns:
            iterable sequence: Iterable sequence of matching documents.
        """

    @abstractmethod
    async def update_one(
        self,
        collection: str,
        query: dict[str, Any],
        update: dict[str, Any],
        upsert: bool = False,
    ) -> dict[str, Any]:
        """Update a single document matching the query.

        Args:
            collection (str): The collection name.
            query (dict): The filter query.
            update (dict): The update document.
            upsert (bool): Whether to insert a new document if no match is found.

        Returns:
            dict: Result of the update operation.
        """

    @abstractmethod
    async def delete_one(self, collection: str, query: dict[str, Any]) -> dict[str, Any]:
        """Delete a single document matching the query.

        Args:
            collection (str): The collection name.
            query (dict): The filter query.

        Returns:
            dict: Result of the delete operation.
        """

    @abstractmethod
    async def count(self, collection: str, query: dict[str, Any]) -> int:
        """Count how many records are in a collection that match the specified query.

        Args:
            collection (str): The collection name.
            query (dict): The filter query.

        Returns:
            int: Number of documents that match the specified query
        """

    @abstractmethod
    async def exists(self, collection: str, query: dict[str, Any]) -> bool:
        """Check if at least one document exists in the collection that matches the specified query.

        Args:
            collection (str): The collection name.
            query (dict): The filter query.

        Returns:
            bool: True if a document exists and False if it doesn't
        """

    @abstractmethod
    async def create_index(self, collection: str, index: tuple[IndexExpression]) -> None:
        """Create an index on a collection in the MongoDB Database.

        Args:
            collection (str): The collection name.
            index (tuple[IndexExpression]):
                A tuple of IndexExpression objects representing the index to create
                NOTE: Muiltiple elements in tuple indicate a single compound index
                not multiple indexes
        """

    @abstractmethod
    async def update_many(
        self,
        collection: str,
        query: dict[str, Any],
        update: dict[str, Any],
    ) -> dict[str, Any]:
        """Update multiple documents matching the query.

        Args:
            collection (str): Name of the collection.
            query (dict): MongoDB filter.
            update (dict): MongoDB update expression.

        Returns:
            dict: Result of the update operation.
        """



import contextvars
from types import TracebackType
from typing import Any

from pydantic import BaseModel

from pydongo.expressions.base import BaseExpression


class MutationExpression(BaseExpression):
    """Expression representing a mutation operation on a document field.

    Args:
        field_name (str): The name of the field to mutate.
        operation (str): The mutation operation (e.g., "$set", "$inc").
        value (Any): The value associated with the mutation operation.
    """

    def __init__(
        self,
        field_name: str,
        operation: str,
        value: Any,
    ):
        self.field_name = field_name
        self.operation = operation
        self.value = value

    def serialize(self) -> dict[str, dict[str, Any]]:
        """Serializes the mutation expression into a MongoDB-style update operation.

        Returns:
            dict: A dictionary representing the mutation operation.
        """
        value = self.value
        if isinstance(value, BaseModel):
            value = value.model_dump()
        return {self.operation: {self.field_name: value}}


class MutationExpressionContext:
    """Context Manager for Mutation expressions across the call stack."""

    _mutations: contextvars.ContextVar[dict[str, Any]] = contextvars.ContextVar("mutations")

    def __enter__(self) -> "MutationExpressionContext":
        """Enter context."""
        self._token = self._mutations.set({})
        return self

    def __exit__(
        self, exc_type: type[BaseException] | None, exc_value: BaseException | None, traceback: TracebackType | None
    ) -> None:
        """Exit current context."""
        self._mutations.reset(self._token)

    @classmethod
    def add_mutation(cls, mutation: "MutationExpression") -> None:
        """Add mutations to context."""
        mutations = cls._mutations.get()
        serialized = mutation.serialize()

        for operator, values in serialized.items():
            mutations.setdefault(operator, {}).update(values)

    @classmethod
    def get_mutations(cls) -> dict[str, Any]:
        """Get mutations from context."""
        return cls._mutations.get()

    @classmethod
    def clear(cls) -> None:
        """Clear the current context."""
        cls._mutations.set({})

    @classmethod
    def has_context(cls) -> bool:
        """Check if current execution has context."""
        try:
            cls._mutations.get()
            return True
        except LookupError:
            return False
